class DatasetExample(object):

    """Sample dataset to demonstrate a simple extension with Cihai.

    This object is passed into :meth:`cihai.Cihai.use` in the
    :class:`cihai.Cihai` object.

    """

    def get(self, request, response):
        """Return chinese character data from sample dataset.

        The source of information in this example is a :obj:`dict`. In real
        modules, the dataset may be a CSV, sqlite database, or an API. As long
        as the data appends to the :param:`response` and returns it.

        :param request: The character or string being looked up, e.g. '好'.
        :type request: str
        :returns: Cihai response dictionary
        :rtype: dict

        """
        dataset = {"好": {"definition": "hao"}}

        if request in dataset:
            response.update(dataset[request])

        return response

    def reverse(self, request, response):
        """Return chinese character data from a reverse lookup sample dataset.

        :param request: The character or string being looked up, e.g. '好'.
        :type request: str
        :returns: Cihai reverse look up. The results should be formatted as::

            {
                '好': {
                    'definition': 'good.'
                }
            }

        When the character already exists in :param:`response`, the character
        ``好`` must be mixed-in with, not overwritten.

        :rtype: dict

        """
        dataset = {"好": {"definition": "hao"}}

        for char in dataset.keys():
            for val in dataset[char].values():
                if request in val:
                    response.update({char: dataset[char]})

        return response



#: Default configuration
DEFAULT_CONFIG = {
    "debug": False,
    "database": {"url": "sqlite:///{user_data_dir}/cihai.db"},
    "dirs": {
        "cache": "{user_cache_dir}",
        "log": "{user_log_dir}",
        "data": "{user_data_dir}",
    },
}

#: User will be prompted to automatically configure their installation for UNIHAN
UNIHAN_CONFIG = {
    "datasets": {"unihan": "cihai.data.unihan.dataset.Unihan"},
    # Turn off by default for using as a plugin example in examples/
    # "plugins": {"unihan": {"variants": "cihai.data.unihan.dataset.UnihanVariants"}},
}



#!/usr/bin/env python
"""
Demonstrate what basic_usage's unihan=True (default Cihai) does under
the hood.
"""
from cihai.core import Cihai


def run(unihan_options=None):
    c = Cihai(unihan=False)
    c.add_dataset("cihai.data.unihan.dataset.Unihan", namespace="unihan")

    if not c.unihan.is_bootstrapped:  # download and install Unihan to db
        c.unihan.bootstrap(unihan_options)

    query = c.unihan.lookup_char("好")
    glyph = query.first()
    print("lookup for 好: %s" % glyph.kDefinition)

    query = c.unihan.reverse_char("good")
    print('matches for "good": %s ' % ", ".join([glph.char for glph in query]))


if __name__ == "__main__":
    run()



import os

from appdirs import AppDirs


def expand_config(d, dirs):
    """
    Expand configuration XDG variables, environmental variables, and tildes.

    Parameters
    ----------
    d : dict
        config information
    dirs : appdirs.AppDirs
        XDG application mapping

    Notes
    -----
    *Environmentable variables* are expanded via :py:func:`os.path.expandvars`.
    So ``${PWD}`` would be replaced by the current PWD in the shell,
    ``${USER}`` would be the user running the app.

    *XDG variables* are expanded via :py:meth:`str.format`. These do not have a
    dollar sign. They are:

    - ``{user_cache_dir}``
    - ``{user_config_dir}``
    - ``{user_data_dir}``
    - ``{user_log_dir}``
    - ``{site_config_dir}``
    - ``{site_data_dir}``

    See Also
    --------
    os.path.expanduser, os.path.expandvars :
        Standard library functions for expanding variables. Same concept, used inside.
    """
    context = {
        "user_cache_dir": dirs.user_cache_dir,
        "user_config_dir": dirs.user_config_dir,
        "user_data_dir": dirs.user_data_dir,
        "user_log_dir": dirs.user_log_dir,
        "site_config_dir": dirs.site_config_dir,
        "site_data_dir": dirs.site_data_dir,
    }

    for k, v in d.items():
        if isinstance(v, dict):
            expand_config(v, dirs)
        if isinstance(v, str):
            d[k] = os.path.expanduser(os.path.expandvars(d[k]))
            d[k] = d[k].format(**context)


class Configurator(object):
    def __init__(self, namespace=""):
        """
        Manage config. Provides facilities for loading / writing configs.

        Used on Cihai and available for its extensions.

        Parameters
        ----------
        namespace : str, optional
            Creates a configuration object that reads/writes on a namespace.

            Leaving this empty / reads and writes to the root level cihai config.

            Namespace is designed for plugins to manage settings.

        Attributes
        ----------
        dirs : appdirs.AppDirs
            XDG App directory locations for cihai

        Class Attributes
        ----------------
        data : dict
            property / getter for options
        _data : dict
            where the raw dictionary resides
        """
        self.dirs = AppDirs("cihai", "cihai team")  # appname  # app author

        self.namespace = namespace

    def get_names(self):
        """
        Return a list of possible places config can reside, and order of search.

        This is based on XDG. So it will look for
        """
        pass

    @property
    def file(self):
        """Find a config file where it exists, as the first place."""
        return

    def read(self):
        """Read to dictionary."""
        return

    def get_delta(self, **updates):
        """Returns the difference of whatever user customizations differ from
        cihai.constants.DEFAULT_CONFIG.
        """
        pass

    def write(self, **updates):
        """If no delta is created from DEFAULT, it not write.

        If file doesn't exist, it will create.
        """
        if updates:
            self._data.update(**updates)
        # save file
        pass



"""
Utility and helper methods for cihai.
"""
import sys

from . import exc


def supports_wide():
    """Return affirmative if python interpreter supports wide characters.

    Returns
    -------
    bool :
        True if python supports wide character sets
    """
    return sys.maxunicode > 0xFFFF


def import_string(import_name: str, silent: bool = False):
    """Imports an object based on a string.

    This is useful if you want to use import paths as endpoints or
    something similar.  An import path can  be specified either in dotted
    notation (``xml.sax.saxutils.escape``) or with a colon as object
    delimiter (``xml.sax.saxutils:escape``).

    If `silent` is True the return value will be `None` if the import fails.

    Parameters
    ----------
    import_name : string
        the dotted name for the object to import.
    silent : bool
        if set to `True` import errors are ignored and `None` is returned instead.

    Returns
    -------
    imported object

    Raises
    ------
    cihai.exc.ImportStringError (ImportError, cihai.exc.CihaiException)

    Notes
    -----
    This is from werkzeug.utils d36aaf1 on May 23, 2022, LICENSE BSD.
    https://github.com/pallets/werkzeug

    Changes:
    - Exception raised is cihai.exc.ImportStringError
    - Format with black
    """
    import_name = import_name.replace(":", ".")
    try:
        try:
            __import__(import_name)
        except ImportError:
            if "." not in import_name:
                raise
        else:
            return sys.modules[import_name]

        module_name, obj_name = import_name.rsplit(".", 1)
        module = __import__(module_name, globals(), locals(), [obj_name])
        try:
            return getattr(module, obj_name)
        except AttributeError as e:
            raise ImportError(e) from None
    except ImportError as e:
        if not silent:
            raise exc.ImportStringError(import_name, e).with_traceback(
                sys.exc_info()[2]
            ) from None
    return None



from cihai import extend
from cihai.core import Cihai
from cihai.data.unihan.dataset import Unihan, UnihanVariants


class SimplestDataset(extend.Dataset):
    def a_method(self):
        return "hi"


def test_add_dataset():
    c = Cihai()
    c.add_dataset(SimplestDataset, namespace="simple")
    assert hasattr(c, "simple")
    assert isinstance(c.simple, extend.Dataset)
    assert hasattr(c.simple, "a_method")
    assert callable(c.simple.a_method)
    assert c.simple.a_method() == "hi"


class SimplestSQLAlchemyDataset(extend.Dataset, extend.SQLAlchemyMixin):
    def a_method(self):
        return "hi"


def test_add_dataset_with_db():
    c = Cihai()
    c.add_dataset(SimplestSQLAlchemyDataset, namespace="simple")
    assert hasattr(c, "simple")
    assert isinstance(c.simple, extend.Dataset)
    assert hasattr(c.simple, "a_method")
    assert callable(c.simple.a_method)
    assert c.simple.a_method() == "hi"

    assert hasattr(c, "sql")
    assert hasattr(c.simple, "sql")


def test_add_dataset_unihan(unihan_options):
    c = Cihai()
    c.add_dataset(Unihan, namespace="unihan")
    assert hasattr(c, "unihan")
    assert isinstance(c.unihan, extend.Dataset)

    c.unihan.sql

    c.unihan.bootstrap(options=unihan_options)
    U = c.sql.base.classes.Unihan

    first_glyph = (
        c.unihan.sql.session.query(U).filter(U.kDefinition.isnot(None)).first()
    )

    char = first_glyph.char
    assert (
        c.unihan.lookup_char(char=char).first().kDefinition == first_glyph.kDefinition
    )

    assert (
        c.unihan.reverse_char(hints=[first_glyph.kDefinition]).first().char == char
    ), "works with list of column value matches"

    assert (
        c.unihan.reverse_char(hints=first_glyph.kDefinition).first().char == char
    ), "works with strings"

    c.unihan.add_plugin(UnihanVariants, "variants")
    assert hasattr(c.unihan, "variants")

    def variant_list(field):
        for char in c.unihan.with_fields(field):
            variants = []
            for var in char.untagged_vars(field):
                variants.append(var)
            yield (char, variants)

    result = {char: variants for (char, variants) in variant_list("kZVariant")}

    assert len(result.values()) > 0
    assert len(result.keys()) > 0



import os
import zipfile

import pytest

from cihai.data.unihan.constants import UNIHAN_FILES


@pytest.fixture
def fixture_path():
    return os.path.abspath(os.path.join(os.path.dirname(__file__), "fixtures"))


@pytest.fixture
def test_config_file(fixture_path):
    return os.path.join(fixture_path, "test_config.yml")


@pytest.fixture
def zip_path(tmpdir):
    return tmpdir.join("Unihan.zip")


@pytest.fixture
def zip_file(zip_path, fixture_path):
    _files = []
    for f in UNIHAN_FILES:
        _files += [os.path.join(fixture_path, f)]
    zf = zipfile.ZipFile(str(zip_path), "a")
    for f in _files:
        zf.write(f, os.path.basename(f))
    zf.close()
    return zf


@pytest.fixture
def unihan_options(zip_file, zip_path, tmpdir):
    return {
        "source": str(zip_path),
        "work_dir": str(tmpdir),
        "zip_path": str(tmpdir.join("downloads").join("Moo.zip")),
    }


@pytest.fixture(scope="function")
def tmpdb_file(tmpdir):
    return tmpdir.join("test.db")



"""Tests for cihai.

Test :class:`Cihai` object. Other tests will use an instance of ``Cihai``
using the ``test_config.yml``.

"""
import cihai
from cihai.constants import UNIHAN_CONFIG
from cihai.core import Cihai
from cihai.data.unihan import bootstrap


def test_cihai_version():
    assert cihai.__version__


def test_config_defaults():
    """Test config defaults."""

    app = Cihai()

    assert "database" in app.config


def test_config_dict_args():
    """Accepts dict as config."""

    expected = "world"

    app = Cihai({"hello": expected})

    result = app.config["hello"]

    assert result == expected


def test_yaml_config_and_override(test_config_file):
    app = Cihai.from_file(test_config_file)

    assert app.config["database"]


def test_unihan_options(unihan_options, test_config_file):
    app = Cihai.from_file(test_config_file)
    bootstrap.bootstrap_unihan(app.sql.metadata, unihan_options)
    assert "Unihan" in app.sql.metadata.tables
    assert app.sql.metadata.tables["Unihan"].columns
    assert set(app.sql.metadata.tables["Unihan"].columns.keys()) == set(
        bootstrap.UNIHAN_FIELDS + ["ucn", "char"]
    )
    assert bootstrap.is_bootstrapped(app.sql.metadata)


def test_bootstraps_unihan_by_default():
    app = Cihai()
    assert UNIHAN_CONFIG.items() == app.config.items()
    assert app.unihan, "cihai bootstraps unihan by default"


def test_cihai_without_unihan():
    app = Cihai(unihan=False)
    assert (
        UNIHAN_CONFIG.items() != app.config.items()
    ), "app can be initialized without unihan"
    assert not hasattr(app, "unihan")



import os

from appdirs import AppDirs

from cihai.config import Configurator, expand_config

#: XDG App directory locations
dirs = AppDirs("cihai", "cihai team")  # appname  # app author


def test_configurator(tmpdir):
    c = Configurator()
    isinstance(c.dirs, AppDirs)

    assert c


def test_expand_config_xdg_vars():
    initial_dict = {
        "dirs": {"cache": "{user_cache_dir}", "data": "{user_cache_dir}/data"}
    }

    expected_dict = {
        "dirs": {
            "cache": dirs.user_cache_dir,
            "data": os.path.join(dirs.user_cache_dir, "data"),
        }
    }

    expand_config(initial_dict, dirs)
    assert initial_dict == expected_dict


def test_expand_config_user_vars():
    initial_dict = {"dirs": {"cache": "~"}}

    expected_dict = {"dirs": {"cache": os.path.expanduser("~")}}

    expand_config(initial_dict, dirs)
    assert initial_dict == expected_dict


def test_expand_config_env_vars(tmpdir, monkeypatch):
    monkeypatch.setenv("MYDIR", str(tmpdir))
    initial_dict = {"dirs": {"cache": "${MYDIR}"}}

    expected_dict = {"dirs": {"cache": os.environ.get("MYDIR")}}

    expand_config(initial_dict, dirs)
    assert initial_dict == expected_dict



#: Mapping of files from unihan-etl (UNIHAN database)
UNIHAN_FILES = [
    "Unihan_DictionaryLikeData.txt",
    "Unihan_IRGSources.txt",
    "Unihan_NumericValues.txt",
    "Unihan_RadicalStrokeCounts.txt",
    "Unihan_Readings.txt",
    "Unihan_Variants.txt",
]


#: Mapping of field names from unihan-etl (UNIHAN database)
UNIHAN_FIELDS = [
    "kAccountingNumeric",
    "kCangjie",
    "kCantonese",
    "kCheungBauer",
    "kCihaiT",
    "kCompatibilityVariant",
    "kDefinition",
    "kFenn",
    "kFourCornerCode",
    "kFrequency",
    "kGradeLevel",
    "kHDZRadBreak",
    "kHKGlyph",
    "kHangul",
    "kHanyuPinlu",
    "kHanyuPinyin",
    "kJapaneseKun",
    "kJapaneseOn",
    "kKorean",
    "kMandarin",
    "kOtherNumeric",
    "kPhonetic",
    "kPrimaryNumeric",
    "kRSAdobe_Japan1_6",
    "kRSJapanese",
    "kRSKanWa",
    "kRSKangXi",
    "kRSKorean",
    "kRSUnicode",
    "kSemanticVariant",
    "kSimplifiedVariant",
    "kSpecializedSemanticVariant",
    "kTang",
    "kTotalStrokes",
    "kTraditionalVariant",
    "kVietnamese",
    "kXHC1983",
    "kZVariant",
]

#: Default settings passed to unihan-etl
UNIHAN_ETL_DEFAULT_OPTIONS = {
    "input_files": UNIHAN_FILES,
    "fields": UNIHAN_FIELDS,
    "format": "python",
    "expand": False,
}



import pytest

from cihai import exc, utils
from unihan_etl.util import merge_dict


def test_merge_dict():
    dict1 = {"hi world": 1, "innerdict": {"hey": 1}}
    dict2 = {"innerdict": {"welcome": 2}}

    expected = {"hi world": 1, "innerdict": {"hey": 1, "welcome": 2}}

    assert merge_dict(dict1, dict2) == expected


def test_import_string():
    utils.import_string("cihai")

    with pytest.raises((ImportError, exc.CihaiException, exc.ImportStringError)):
        utils.import_string("cihai.core.nonexistingimport")

    with pytest.raises((ImportError, exc.CihaiException, exc.ImportStringError)):
        utils.import_string("cihai2")



#!/usr/bin/env python
from cihai.core import Cihai


def variant_list(unihan, field):
    for char in unihan.with_fields(field):
        print("Character: {}".format(char.char))
        for var in char.untagged_vars(field):
            print(var)


def run(unihan_options=None):
    """Wrapped so we can test in tests/test_examples.py"""
    print("This example prints variant character data.")

    c = Cihai()
    if not c.unihan.is_bootstrapped:  # download and install Unihan to db
        c.unihan.bootstrap(unihan_options)

    c.unihan.add_plugin(
        "cihai.data.unihan.dataset.UnihanVariants", namespace="variants"
    )

    print("## ZVariants")
    variant_list(c.unihan, "kZVariant")

    print("## kSemanticVariant")
    variant_list(c.unihan, "kSemanticVariant")

    print("## kSpecializedSemanticVariant")
    variant_list(c.unihan, "kSpecializedSemanticVariant")


if __name__ == "__main__":
    run()



from sqlalchemy import Column, or_

from ...conversion import parse_untagged, parse_vars
from ...extend import Dataset, DatasetPlugin, SQLAlchemyMixin
from . import bootstrap


class Unihan(Dataset, SQLAlchemyMixin):
    def bootstrap(self, options=None):
        if options is None:
            options = {}

        bootstrap.bootstrap_unihan(self.sql.metadata, options=options)
        self.sql.reflect_db()  # automap new table created during bootstrap

    def lookup_char(self, char):
        """Return character information from datasets.

        Parameters
        ----------
        char : str
            character / string to lookup

        Returns
        -------
        :class:`sqlalchemy.orm.query.Query` :
            list of matches
        """
        Unihan = self.sql.base.classes.Unihan
        return self.sql.session.query(Unihan).filter_by(char=char)

    def reverse_char(self, hints):
        """Return QuerySet of objects from SQLAlchemy of results.

        Parameters
        ----------
        hints: list of str
            strings to lookup

        Returns
        -------
        :class:`sqlalchemy.orm.query.Query` :
            reverse matches
        """
        if isinstance(hints, str):
            hints = [hints]

        Unihan = self.sql.base.classes.Unihan
        columns = Unihan.__table__.columns
        return self.sql.session.query(Unihan).filter(
            or_(*[column.contains(hint) for column in columns for hint in hints])
        )

    def with_fields(self, *fields):
        """Returns list of characters with information for certain fields.

        Parameters
        ----------
        *fields : list of str
            fields for which information should be available

        Returns
        -------
        :class:`sqlalchemy.orm.query.Query` :
            list of matches
        """
        Unihan = self.sql.base.classes.Unihan
        query = self.sql.session.query(Unihan)
        for field in fields:
            query = query.filter(Column(field).isnot(None))
        return query

    @property
    def is_bootstrapped(self):
        """Return True if UNIHAN and database is set up.

        Returns
        -------
        bool :
            True if Unihan application fixture data installed.
        """
        return bootstrap.is_bootstrapped(self.sql.metadata)


class UnihanVariants(DatasetPlugin):
    def bootstrap(self):
        def tagged_vars(table, col):
            """
            Return a variant column as an iterator of (char, tag) tuples.
            """
            return parse_vars(getattr(table, col))

        def untagged_vars(table, col):
            """
            Return a variant column as an iterator of chars.
            """
            return parse_untagged(getattr(table, col))

        if hasattr(self.sql.base.classes, "Unihan"):
            self.sql.base.classes.Unihan.tagged_vars = tagged_vars
            self.sql.base.classes.Unihan.untagged_vars = untagged_vars



# flake8: noqa: E501
"""Conversion functions for various CJK encodings and representations.

Notes
-----
Original methods and docs based upon `ltchinese`_, license `MIT`_ Steven
Kryskalla.

.. versionadded:: 0.1
    Python 2/3 compatibility.

    - PEP8, PEP257.
    - ``int()`` casting for comparisons
    - Python 3 support.
    - Python 3 fix for :meth:`~.ucn_to_python`.
    - Python 3 ``__future__`` statements.
    - All methods converting to ``_python`` will return ``Unicode``.
    - All methods converting Unicode to x will return bytestring.
    - Add :meth:`~.ucnstring_to_python`
    - Any other change upon @ `conversion.py @9227813`_.

The following terms are used to represent the encodings / representation used
in the conversion functions (the samples on the right are for the character
U+4E00 (yi1; "one")):

========================== ====================================================
GB2312 (Kuten/Quwei form)  "5027" [used in the "GB2312" field of Unihan.txt]
GB2312 (ISO-2022 form)     "523B" [the "internal representation" of GB code]
EUC-CN                     "D2BB" [this is the "external encoding" of GB2312-
                                    ISO2022's "internal representation"; also
                                    the form that Ocrat uses]
UTF-8                      "E4 B8 80" [used in the "UTF-8" field in Unihan.txt]
-------------------------- ----------------------------------------------------
Unihan UCN                 "U+4E00"   [used by Unicode Inc.]
-------------------------- ----------------------------------------------------
internal Python unicode    u"\u4e00"  [this is the most useful form!]
internal Python 'utf8'     "\\xe4\\xb8\\x80"
internal Python 'gb2312'   "\\xd2\\xbb"
internal Python 'euc-cn'   "\\xd2\\xbb"
internal Python 'gb18030'  "\\xd2\\xbb"
========================== ====================================================

See these resources for more information:
 * Wikipedia "Extended_Unix_Code" article

   * "EUC-CN is the usual way to use the GB2312 standard for simplified Chinese
     characters ... the ISO-2022 form of GB2312 is not normally used"

 * Wikipedia "HZ_(encoding)" article (the example conversion)

 * Wikipedia "Numeric_character_reference" article

 * Unihan (look for "Encoding forms", "Mappings to Major Standards")

   * e.g. http://www.unicode.org/cgi-bin/GetUnihanData.pl?codepoint=4E00

.. _ltchinese: https://bitbucket.org/lost_theory/ltchinese
.. _MIT: https://bitbucket.org/lost_theory/ltchinese/src/9227813/LICENSE.txt
.. _conversion.py @9227813: https://bitbucket.org/lost_theory/ltchinese/raw/9227813/ltchinese/conversion.py
"""
import logging
import re

log = logging.getLogger(__name__)


def hexd(n):
    """Return hex digits (strip '0x' at the beginning)."""
    return hex(n)[2:]


def kuten_to_gb2312(kuten):
    """
    Convert GB kuten / quwei form (94 zones * 94 points) to GB2312-1980 /
    ISO-2022-CN hex (internal representation)
    """
    zone, point = int(kuten[:2]), int(kuten[2:])
    hi, lo = hexd(zone + 0x20), hexd(point + 0x20)

    gb2312 = "%s%s" % (hi, lo)

    assert isinstance(gb2312, bytes)
    return gb2312


def gb2312_to_euc(gb2312hex):
    """
    Convert GB2312-1980 hex (internal representation) to EUC-CN hex (the
    "external encoding")
    """
    hi, lo = int(gb2312hex[:2], 16), int(gb2312hex[2:], 16)
    hi, lo = hexd(hi + 0x80), hexd(lo + 0x80)

    euc = "%s%s" % (hi, lo)
    assert isinstance(euc, bytes)
    return euc


def euc_to_python(hexstr):
    """
    Convert a EUC-CN (GB2312) hex to a Python unicode string.
    """
    hi = hexstr[0:2]
    lo = hexstr[2:4]
    gb_enc = b"\\x" + hi + b"\\x" + lo
    return gb_enc.decode("gb2312")


def euc_to_utf8(euchex):
    """
    Convert EUC hex (e.g. "d2bb") to UTF8 hex (e.g. "e4 b8 80").
    """
    utf8 = euc_to_python(euchex).encode("utf-8")
    uf8 = utf8.decode("unicode_escape")

    uf8 = uf8.encode("latin1")

    uf8 = uf8.decode("euc-jp")
    return uf8


def ucn_to_unicode(ucn):
    """
    Convert a Unicode Universal Character Number (e.g. "U+4E00" or "4E00") to
    Python unicode (u'\\u4e00')
    """
    if isinstance(ucn, str):
        ucn = ucn.strip("U+")
        if len(ucn) > int(4):
            char = b"\\U" + format(int(ucn, 16), "08x").encode("latin1")
            char = char.decode("unicode_escape")
        else:
            char = chr(int(ucn, 16))
    else:
        char = chr(ucn)

    assert isinstance(char, str)

    return char


def euc_to_unicode(hexstr):
    r"""
    Return EUC-CN (GB2312) hex to a Python unicode.

    Parameters
    ----------
    hexstr : bytes

    Returns
    -------
    unicode :
        Python unicode  e.g. ``u'\\u4e00'`` / '一'.

    Examples
    --------

    >>> u'\u4e00'.encode('gb2312').decode('utf-8')
    '\u04bb'

    >>> (b'\\x' + b'd2' + b'\\x' + b'bb').replace('\\x', '') \  # doctest: +SKIP
    ... .decode('hex').decode('utf-8')
    u'\u04bb'

    Note: bytes don't have a ``.replace``:

    >>> gb_enc = gb_enc.replace('\\x', '').decode('hex')  # doctest: +SKIP
    >>> gb_enc.decode('string_escape')  # Won't work with Python 3.x.  # doctest: +SKIP
    """
    hi = hexstr[0:2]
    lo = hexstr[2:4]
    # hi and lo are only 2 characters long, no risk with eval-ing them

    gb_enc = b"\\x" + hi + b"\\x" + lo
    assert isinstance(gb_enc, bytes)

    # Requires coercing back to str in 2.7
    gb_enc = gb_enc.decode("unicode_escape")

    gb_enc = gb_enc.encode("latin1")

    gb_enc = gb_enc.decode("gb2312")

    assert isinstance(gb_enc, str)
    return gb_enc


""" Convert from internal Python unicode / string objects """


def python_to_ucn(uni_char, as_bytes=False):
    """
    Return UCN character from Python Unicode character.

    Converts a one character Python unicode string (e.g. u'\\u4e00') to the
    corresponding Unicode UCN ('U+4E00').
    """
    ucn = uni_char.encode("unicode_escape").decode("latin1")
    ucn = str(ucn).replace("\\", "").upper().lstrip("U")
    if len(ucn) > int(4):
        # get rid of the zeroes that Python uses to pad 32 byte UCNs
        ucn = ucn.lstrip("0")
    ucn = "U+" + ucn.upper()

    if as_bytes:
        ucn = ucn.encode("latin1")

    return ucn


def python_to_euc(uni_char, as_bytes=False):
    """
    Return EUC character from a Python Unicode character.

    Converts a one character Python unicode string (e.g. u'\\u4e00') to the
    corresponding EUC hex ('d2bb').
    """
    euc = repr(uni_char.encode("gb2312"))[1:-1].replace("\\x", "").strip("'")

    if as_bytes:
        euc = euc.encode("utf-8")
        assert isinstance(euc, bytes)

    return euc


def ucnstring_to_unicode(ucn_string):
    """Return ucnstring as Unicode."""
    ucn_string = ucnstring_to_python(ucn_string).decode("utf-8")

    assert isinstance(ucn_string, str)
    return ucn_string


def ucnstring_to_python(ucn_string):
    """
    Return string with Unicode UCN (e.g. "U+4E00") to native Python Unicode
    (u'\\u4e00').
    """
    res = re.findall(r"U\+[0-9a-fA-F]*", ucn_string)
    for r in res:
        ucn_string = ucn_string.replace(str(r), str(ucn_to_unicode(r)))

    ucn_string = ucn_string.encode("utf-8")

    assert isinstance(ucn_string, bytes)
    return ucn_string


def parse_var(var):
    """
    Returns a tuple consisting of a string and a tag, or None, if none is
    specified.
    """
    bits = var.split("<", 1)
    if len(bits) < 2:
        tag = None
    else:
        tag = bits[1]
    return ucn_to_unicode(bits[0]), tag


def parse_vars(_vars):
    """
    Return an iterator of (char, tag) tuples.
    """
    for var in _vars.split(" "):
        yield parse_var(var)


def parse_untagged(_vars):
    """
    Return an iterator of chars.
    """
    return (char for char, _tag in parse_vars(_vars))



"""
Cihai Plugin System

Status: Experimental, API can change

As a pilot, the UNIHAN library, and an plugin for it, in #131 [1]_

You can bring any data layout / backend you like to cihai.

For convenience, you can use cihai's configuration namespace and SQLAlchemy settings.

You can also create plugins which extend another. So if Unihan doesn't have a lookup
for variant glyphs, this can be added.
"""
from . import utils


class ConfigMixin(object):
    """
    This piggybacks cihai's global config state, as well as your datasets.

    Cihai will automatically manage the user's config, as well as your datasets,
    neatly in XDG.

    Raises
    ------
    Functions inside, and what you write relating to dataset config should return

    CihaiDatasetConfigException (CihaiDatasetException)

    config.cihai = links directly back to Cihai's configuration dictionary
    (todo note: make this non-mutable property)

    config : dict
        your local user's config

    check() : function, optional
        this is ran on start. it can raise DatasetConfigException

    default_config : your dataset's default configuration

    get_default_config : override function in case you'd like custom configs (for
        instnace if you want a platform to use a different db driver, or do version
        checks, etc.)

        internal functions use get_default_config()
    """


class SQLAlchemyMixin(object):
    """Your dataset can use any backend you'd like, we provide a backend for you, that
    automatically piggybacks on cihai's zero-config, XDG / SQLAchemy configuration. So
    it's preconfigured for the user.

    In addition, this mixin gives you access to any other of the user's sqlalchemy
    sql that use this mixin. So if you want a dataset that utilitizes UNIHAN, you can
    access that easily.

    This will provide the following instance-level properties in methods:

    When you have access, it's expected to keep your tables / databases namespaced so
    they don't clobber.
    """

    #: :class:`sqlalchemy.engine.Engine` instance.
    engine = None

    #: :class:`sqlalchemy.schema.MetaData` instance.
    metadata = None

    #: :class:`sqlalchemy.orm.session.Session` instance.
    session = None

    #: :class:`sqlalchemy.ext.automap.AutomapBase` instance.
    base = None


class Dataset(object):
    """
    Cihai dataset, e.g. UNIHAN.

    See Also
    --------
    cihai.data.unihan.dataset.Unihan : reference implementation
    """

    def bootstrap(self):
        pass

    def add_plugin(self, _cls, namespace, bootstrap=True):
        if isinstance(_cls, str):
            _cls = utils.import_string(_cls)
        setattr(self, namespace, _cls())
        plugin = getattr(self, namespace)

        if hasattr(self, "sql") and isinstance(self, SQLAlchemyMixin):
            plugin.sql = self.sql

        if bootstrap and hasattr(plugin, "bootstrap") and callable(plugin.bootstrap):
            plugin.bootstrap()


class DatasetPlugin(object):
    """
    Extend the functionality of datasets with custom methods, actions, etc.

    See Also
    --------
    cihai.data.unihan.dataset.UnihanVariants : reference implementation
    """



"""Cihai core functionality."""
import logging
import os

import kaptan
from appdirs import AppDirs

from unihan_etl.util import merge_dict

from . import exc, extend
from .config import expand_config
from .constants import DEFAULT_CONFIG, UNIHAN_CONFIG
from .db import Database
from .utils import import_string

log = logging.getLogger(__name__)


class Cihai(object):
    """
    Central application object.

    By default, this automatically adds the UNIHAN dataset.

    Attributes
    ----------
    config : dict

    Notes
    -----
    Inspired by the early pypa/warehouse applicaton object [1]_.

    **Configuration templates**

    The ``config`` :py:class:`dict` parameter supports a basic template system
    for replacing :term:`XDG Base Directory` directory variables, tildes
    and environmentas variables. This is done by passing the option dict
    through :func:`cihai.config.expand_config` during initialization.

    Examples
    --------
    To use cihai programatically, invoke and install the UNIHAN [2]_ dataset:

    .. literalinclude:: ../examples/basic_usage.py
        :language: python

    Above: :attr:`~cihai.data.unihan.bootstrap.is_bootstrapped` can check if the system
    has the database installed.

    References
    ----------
    .. [1] PyPA Warehouse on GitHub. https://github.com/pypa/warehouse.
       Accessed sometime in 2013.
    .. [2] UNICODE HAN DATABASE (UNIHAN) documentation.
       https://www.unicode.org/reports/tr38/. Accessed March 31st, 2018.
    """

    #: :py:class:`dict` of default config, can be monkey-patched during tests
    default_config = DEFAULT_CONFIG

    def __init__(self, config=None, unihan=True):
        """
        Parameters
        ----------
        config : dict, optional
        unihan : boolean, optional
            Bootstrap the core UNIHAN dataset (recommended)
        """
        if config is None:
            config = {}

        # Merges custom configuration settings on top of defaults
        #: Configuration dictionary
        self.config = merge_dict(self.default_config, config)

        if unihan:
            self.config = merge_dict(UNIHAN_CONFIG, self.config)

        #: XDG App directory locations
        dirs = AppDirs("cihai", "cihai team")  # appname  # app author

        #: Expand template variables
        expand_config(self.config, dirs)

        if not os.path.exists(dirs.user_data_dir):
            os.makedirs(dirs.user_data_dir)

        #: :class:`cihai.db.Database` : Database instance
        self.sql = Database(self.config)

        self.bootstrap()

    def bootstrap(self):
        for namespace, class_string in self.config.get("datasets", {}).items():
            self.add_dataset(class_string, namespace)

        for dataset, plugins in self.config.get("plugins", {}).items():
            for namespace, class_string in plugins.items():
                getattr(self, dataset).add_plugin(class_string, namespace)

    def add_dataset(self, _cls, namespace):
        if isinstance(_cls, str):
            _cls = import_string(_cls)

        setattr(self, namespace, _cls())
        dataset = getattr(self, namespace)

        if isinstance(dataset, extend.SQLAlchemyMixin):
            dataset.sql = self.sql

    @classmethod
    def from_file(cls, config_path=None, *args, **kwargs):
        """
        Create a Cihai instance from a JSON or YAML config.

        Parameters
        ----------
        config_path : str, optional
            path to custom config file

        Returns
        -------
        :class:`Cihai` :
            application object
        """

        config_reader = kaptan.Kaptan()

        config = {}

        if config_path:
            if not os.path.exists(config_path):
                raise exc.CihaiException(
                    "{0} does not exist.".format(os.path.abspath(config_path))
                )
            if not any(
                config_path.endswith(ext) for ext in ("json", "yml", "yaml", "ini")
            ):
                raise exc.CihaiException(
                    "{0} does not have a yaml,yml,json,ini extend.".format(
                        os.path.abspath(config_path)
                    )
                )
            else:
                custom_config = config_reader.import_config(config_path).get()
                config = merge_dict(config, custom_config)

        return cls(config)



import importlib
import importlib.util
import sys
import types

import pytest


def load_script(example: str) -> types.ModuleType:
    file_path = f"examples/{example}.py"
    module_name = "run"

    spec = importlib.util.spec_from_file_location(module_name, file_path)
    assert spec is not None
    module = importlib.util.module_from_spec(spec)
    sys.modules[module_name] = module

    assert spec.loader is not None
    spec.loader.exec_module(module)

    return module


def test_dataset(unihan_options):
    example = load_script("dataset")
    example.run()


def test_variants(unihan_options):
    example = load_script("variants")
    example.run()


def test_ts_difficulties(unihan_options):
    example = load_script("variant_ts_difficulties")
    example.run(unihan_options=unihan_options)


def test_basic_usage(unihan_options, capsys: pytest.CaptureFixture[str]):
    example = load_script("basic_usage")
    example.run(unihan_options=unihan_options)

    captured = capsys.readouterr()

    assert "lookup for 好: good" in captured.out
    assert 'matches for "good": 好' in captured.out


def test_basic_usage_manual(unihan_options, capsys: pytest.CaptureFixture[str]):
    example = load_script("basic_usage_manual")
    example.run(unihan_options=unihan_options)

    captured = capsys.readouterr()

    assert "lookup for 好: good" in captured.out
    assert 'matches for "good": 好' in captured.out



"""Cihai core functionality."""
from sqlalchemy import MetaData, create_engine
from sqlalchemy.ext.automap import automap_base
from sqlalchemy.orm import Session


class Database(object):
    """
    Cihai SQLAlchemy instance
    """

    def __init__(self, config):
        self.engine = create_engine(config["database"]["url"])

        self.metadata = MetaData()
        self.metadata.bind = self.engine
        self.reflect_db()

        self.session = Session(self.engine)

    def reflect_db(self):
        """
        No-op to reflect db info.

        This is available as a method so the database can be reflected
        outside initialization (such bootstrapping unihan during CLI usage).
        """
        self.metadata.reflect(views=True, extend_existing=True)
        self.base = automap_base(metadata=self.metadata)
        self.base.prepare()

    #: :class:`sqlalchemy.engine.Engine` instance.
    engine = None

    #: :class:`sqlalchemy.schema.MetaData` instance.
    metadata = None

    #: :class:`sqlalchemy.orm.session.Session` instance.
    session = None

    #: :class:`sqlalchemy.ext.automap.AutomapBase` instance.
    base = None



"""Tests for cihai datasets

tests.datasets
~~~~~~~~~~~~~~

These tests will be tested against a fixture of :class:`Cihai` which uses the
settings found in ``test_config.yml``.

"""
import random

import pytest

import sqlalchemy
from sqlalchemy import MetaData

from cihai import conversion

cjk_ranges = {  # http://www.unicode.org/reports/tr38/#BlockListing
    "CJK Unified Ideographs": range(0x4E00, 0x9FD5 + 1),
    "CJK Unified Ideographs Extension A": range(0x3400, 0x4DBF + 1),
    "CJK Unified Ideographs Extension B": range(0x20000, 0x2A6DF + 1),
    "CJK Unified Ideographs Extension C": range(0x2A700, 0x2B73F + 1),
    "CJK Unified Ideographs Extension D": range(0x2B840, 0x2B81F + 1),
    "CJK Unified Ideographs Extension E": range(0x2B820, 0x2CEAF + 1),
    "CJK Radicals Supplement": range(0x2E80, 0x2EFF + 1),
    "CJK Symbols and Punctuation": range(0x3000, 0x303F + 1),
    "CJK Strokes": range(0x31C0, 0x31EF + 1),
    "Ideographic Description Characters": range(0x2FF0, 0x2FFF + 1),
    "Kangxi Radicals": range(0x2F00, 0x2FDF + 1),
    "Enclosed CJK Letters and Months": range(0x3200, 0x32FF + 1),
    "CJK Compatibility": range(0x3300, 0x33FF + 1),
    "CJK Compatibility Ideographs": range(0xF900, 0xFAFF + 1),
    "CJK Compatibility Ideographs Supplement": range(0x2F800, 0x2FA1D + 1),
    "CJK Compatibility Forms": range(0xFE30, 0xFE4F + 1),
    "Yijing Hexagram Symbols": range(0x4DC0, 0x4DFF + 1),
}


engine = sqlalchemy.create_engine("sqlite:///")
metadata = MetaData(bind=engine)

unicode_table = sqlalchemy.Table(
    "cjk",
    metadata,
    sqlalchemy.Column("id", sqlalchemy.Integer(), primary_key=True),
    sqlalchemy.Column("char", sqlalchemy.Unicode()),
    sqlalchemy.Column("ucn", sqlalchemy.String()),
)

sample_table = sqlalchemy.Table(
    "sample_table",
    metadata,
    sqlalchemy.Column("id", sqlalchemy.Integer(), primary_key=True),
    sqlalchemy.Column("char_id", sqlalchemy.ForeignKey("cjk.id")),
    sqlalchemy.Column("value", sqlalchemy.Unicode()),
)

metadata.create_all()


def get_char_fk(char):
    return (
        unicode_table.select(unicode_table.c.id)
        .where(unicode_table.c.char == char)
        .limit(1)
        .execute()
        .fetchone()
        .id
    )


def get_char_fk_multiple(*args):
    """Retrieve the Rows"""

    where_opts = []

    for arg in args:
        where_opts.append(unicode_table.c.char == arg)

    where_opts = sqlalchemy.or_(*where_opts)

    results = unicode_table.select().where(where_opts).execute()

    return results


@pytest.fixture(scope="session")
def chars():
    chars = []

    while len(chars) < 3:
        c = 0x4E00 + random.randint(1, 333)
        char = {
            "hex": c,
            "char": chr(int(c)),
            "ucn": conversion.python_to_ucn(chr(int(c))),
        }
        if char not in chars:
            chars.append(char)

        metadata.bind.execute(unicode_table.insert(), chars)
    return chars


def test_insert_row(chars):

    cjkchar = chars[0]

    row = unicode_table.select().limit(1).execute().fetchone()

    assert row.char == cjkchar["char"]


def test_insert_bad_fk():
    wat = sample_table.insert().values(value="", char_id="wat").execute()

    assert wat


def test_insert_on_foreign_key(chars):

    cjkchar = chars[0]
    char = cjkchar["char"]

    sample_table.insert().values(char_id=get_char_fk(char), value="hey").execute()

    select_char = unicode_table.select().where(unicode_table.c.char == char).limit(1)
    row = select_char.execute().fetchone()

    assert row is not None


def test_get_char_foreign_key_multiple(chars):
    char_fk_multiple = get_char_fk_multiple(*[c["char"] for c in chars])

    for char in char_fk_multiple:
        assert char["char"]



"""Exceptions raised from the Cihai library."""


class CihaiException(Exception):
    """Base Cihai Exception class."""


class ImportStringError(ImportError, CihaiException):
    """
    Provides information about a failed :func:`import_string` attempt.

    Notes
    -----
    This is from werkzeug.utils d36aaf1 on August 20 2022, LICENSE BSD.
    https://github.com/pallets/werkzeug

    Changes:
    - Deferred load import import_string from cihai.util
    - Format with black
    """

    #: String in dotted notation that failed to be imported.
    import_name: str
    #: Wrapped exception.
    exception: BaseException

    def __init__(self, import_name, exception):
        from .utils import import_string

        self.import_name = import_name
        self.exception = exception
        msg = import_name
        name = ""
        tracked = []
        for part in import_name.replace(":", ".").split("."):
            name = f"{name}.{part}" if name else part
            imported = import_string(name, silent=True)
            if imported:
                tracked.append((name, getattr(imported, "__file__", None)))
            else:
                track = [f"- {n!r} found in {i!r}." for n, i in tracked]
                track.append(f"- {name!r} not found.")
                track_str = "\n".join(track)
                msg = (
                    f"import_string() failed for {import_name!r}. Possible reasons"
                    f" are:\n\n"
                    "- missing __init__.py in a package;\n"
                    "- package or module path not included in sys.path;\n"
                    "- duplicated package or module name taking precedence in"
                    " sys.path;\n"
                    "- missing module, class, function or variable;\n\n"
                    f"Debugged import:\n\n{track_str}\n\n"
                    f"Original exception:\n\n{type(exception).__name__}: {exception}"
                )
                break

        super().__init__(msg)

    def __repr__(self) -> str:
        return f"<{type(self).__name__}({self.import_name!r}, {self.exception!r})>"



#!/usr/bin/env python
from cihai.core import Cihai


def run(unihan_options=None):
    c = Cihai()
    if not c.unihan.is_bootstrapped:  # download and install Unihan to db
        c.unihan.bootstrap(unihan_options)

    c.unihan.add_plugin(
        "cihai.data.unihan.dataset.UnihanVariants", namespace="variants"
    )

    print(
        "This example prints some tricky cases of character-by-character "
        "Traditional-Simplified mapping."
    )
    print("https://www.unicode.org/reports/tr38/#N10211")
    print("3.7.1 bullet 4")

    for char in c.unihan.with_fields("kTraditionalVariant", "kSimplifiedVariant"):
        print("Character: {}".format(char.char))
        trad = set(char.untagged_vars("kTraditionalVariant"))
        simp = set(char.untagged_vars("kSimplifiedVariant"))
        Unihan = c.sql.base.classes.Unihan
        if Unihan.char in trad and Unihan.char in simp:
            print("Case 1")
        else:
            print("Case 2 (non-idempotent)")
        for trad_var in trad:
            print("s2t: {}".format(trad_var))
        for simp_var in simp:
            print("t2s: {}".format(simp_var))


if __name__ == "__main__":
    run()



"""Tests for cihai.

test.conversion
~~~~~~~~~~~~~~~

"""
from cihai import conversion


def test_str():
    c1 = "(same as U+7A69 穩) firm; stable; secure"
    c2 = str()

    assert isinstance(c1, str)
    assert isinstance(c2, str)


"""Return UCN character from Python Unicode character.

Converts a one character Python unicode string (e.g. u'\\u4e00') to the
corresponding Unicode UCN ('U+4E00').

U+369D	kSemanticVariant	U+595E<kMatthews U+594E<kMatthews
U+3CE2	kTraditionalVariant	U+23FB7
U+3FF7	kSemanticVariant	U+7CD9<kMatthews,kMeyerWempe
U+345A	kDefinition	(non-classical form of 那) that, there
U+349A	kDefinition	(same as U+7A69 穩) firm; stable; secure,
        dependent upon others
U+34B5	kMandarin	mào
U+356D	kCantonese	au3 jaau1
"""


def test_ucn_from_unicode():
    text = "一"
    python_unicode = "\u4e00"

    expected = "U+4E00"
    bytes_expected = b"U+4E00"

    assert conversion.python_to_ucn(python_unicode) == expected
    assert isinstance(conversion.python_to_ucn(python_unicode), str)
    assert isinstance(conversion.python_to_ucn(python_unicode, as_bytes=True), bytes)

    assert conversion.python_to_ucn(text, as_bytes=True) == bytes_expected


def test_ucn_from_unicode_16():
    text = "𦄀"
    python_unicode = "\u26100"

    expected = "U+26100"
    bytes_expected = b"U+26100"

    assert conversion.python_to_ucn(python_unicode) == expected
    assert isinstance(conversion.python_to_ucn(python_unicode), str)
    assert isinstance(conversion.python_to_ucn(python_unicode, as_bytes=True), bytes)

    assert conversion.python_to_ucn(text, as_bytes=True) == bytes_expected


def test_ucn_to_unicode():
    before = "U+4E00"
    expected = "\u4e00"

    result = conversion.ucn_to_unicode(before)

    assert result == expected

    assert isinstance(result, str)

    # wide character
    before = "U+20001"
    expected = "\U00020001"

    result = conversion.ucn_to_unicode(before)

    assert result == expected
    assert isinstance(result, str)

    before = "(same as U+7A69 穩) firm; stable; secure"
    expected = "(same as 穩 穩) firm; stable; secure"

    result = conversion.ucnstring_to_unicode(before)

    assert result == expected
    assert isinstance(result, str)


"""Return EUC character from a Python Unicode character.

    Converts a one character Python unicode string (e.g. u'\\u4e00') to the
    corresponding EUC hex ('d2bb').

"""


def test_hexd():
    assert conversion.hexd(0xFFFF) == "ffff"


def test_euc_from_unicode():
    expected = "一"  # u'\u4e00'
    euc_bytestring = b"d2bb"
    euc_unicode = "d2bb"

    result = conversion.python_to_euc(expected, as_bytes=True)

    assert euc_bytestring == result
    assert isinstance(result, bytes)

    result = conversion.python_to_euc(expected)

    assert euc_unicode == result
    assert isinstance(result, str)


def test_euc_to_utf8():
    expected = "一"
    euc_bytestring = b"b0ec"

    result = conversion.euc_to_utf8(euc_bytestring)

    assert expected == result


def test_euc_to_unicode():
    expected = "一"
    expected_ustring = "\u4e00"
    euc_bytestring = b"d2bb"

    result = conversion.euc_to_unicode(euc_bytestring)

    assert expected == expected_ustring
    assert isinstance(result, str)

    assert expected == result
    assert expected_ustring == result



from sqlalchemy import Column, String, Table

from unihan_etl import process as unihan
from unihan_etl.process import UNIHAN_MANIFEST
from unihan_etl.util import merge_dict

from .constants import UNIHAN_ETL_DEFAULT_OPTIONS, UNIHAN_FIELDS


def bootstrap_unihan(metadata, options=None):
    if options is None:
        options = {}

    """Download, extract and import unihan to database."""
    options = merge_dict(UNIHAN_ETL_DEFAULT_OPTIONS.copy(), options)

    p = unihan.Packager(options)
    p.download()
    data = p.export()
    table = create_unihan_table(UNIHAN_FIELDS, metadata)
    metadata.create_all()
    metadata.bind.execute(table.insert(), data)


TABLE_NAME = "Unihan"


def flatten_datasets(d):
    return sorted({c for cs in d.values() for c in cs})


DEFAULT_COLUMNS = ["ucn", "char"]
try:
    DEFAULT_FIELDS = [f for t, f in UNIHAN_MANIFEST.items() if t in ["Unihan"]]
except Exception:
    DEFAULT_FIELDS = [f for t, f in UNIHAN_MANIFEST.items()]


def is_bootstrapped(metadata):
    """Return True if cihai is correctly bootstrapped."""
    fields = UNIHAN_FIELDS + DEFAULT_COLUMNS
    if TABLE_NAME in metadata.tables.keys():
        table = metadata.tables[TABLE_NAME]

        if set(fields) == {c.name for c in table.columns}:
            return True
        else:
            return False
    else:
        return False


def create_unihan_table(columns, metadata):
    """Create table and return  :class:`sqlalchemy.Table`.

    Parameters
    ----------
    columns : list
        columns for table, e.g. ``['kDefinition', 'kCantonese']``
    metadata : :class:`sqlalchemy.schema.MetaData`
        Instance of sqlalchemy metadata

    Returns
    -------
    :class:`sqlalchemy.schema.Table` :
        Newly created table with columns and index.
    """

    if TABLE_NAME not in metadata.tables:
        table = Table(TABLE_NAME, metadata)

        table.append_column(Column("char", String(12), primary_key=True))
        table.append_column(Column("ucn", String(12), primary_key=True))

        for column_name in columns:
            col = Column(column_name, String(256), nullable=True)
            table.append_column(col)

        return table
    else:
        return Table(TABLE_NAME, metadata)



"""Log utilities for cihai.

cihai.log
~~~~~~~~~

"""
import logging
import time

from colorama import Fore, Style, init

LEVEL_COLORS = {
    "DEBUG": Fore.BLUE,  # Blue
    "INFO": Fore.GREEN,  # Green
    "WARNING": Fore.YELLOW,
    "ERROR": Fore.RED,
    "CRITICAL": Fore.RED,
}


def default_log_template(self, record):
    """Return the prefix for the log message. Template for Formatter.

    :param: record: :py:class:`logging.LogRecord` object. this is passed in
    from inside the :py:meth:`logging.Formatter.format` record.

    """

    reset = Style.RESET_ALL
    levelname = [
        LEVEL_COLORS.get(record.levelname),
        Style.BRIGHT,
        "(%(levelname)s)",
        Style.RESET_ALL,
        " ",
    ]
    asctime = [
        "[",
        Fore.BLACK,
        Style.DIM,
        Style.BRIGHT,
        "%(asctime)s",
        Fore.RESET,
        Style.RESET_ALL,
        "]",
    ]
    name = [
        " ",
        Fore.WHITE,
        Style.DIM,
        Style.BRIGHT,
        "%(name)s",
        Fore.RESET,
        Style.RESET_ALL,
        " ",
    ]

    tpl = "".join(reset + levelname + asctime + name + reset)

    return tpl


class LogFormatter(logging.Formatter):

    template = default_log_template

    def __init__(self, color=True, *args, **kwargs):
        init()
        logging.Formatter.__init__(self, *args, **kwargs)

    def format(self, record):
        try:
            record.message = record.getMessage()
        except Exception as e:
            record.message = "Bad message (%r): %r" % (e, record.__dict__)

        date_format = "%H:%m:%S"
        record.asctime = time.strftime(date_format, self.converter(record.created))

        prefix = self.template(record) % record.__dict__

        formatted = prefix + " " + record.message
        return formatted.replace("\n", "\n    ")


def debug_log_template(self, record):
    """Return the prefix for the log message. Template for Formatter.

    :param: record: :py:class:`logging.LogRecord` object. this is passed in
    from inside the :py:meth:`logging.Formatter.format` record.

    """

    reset = Style.RESET_ALL
    levelname = [
        LEVEL_COLORS.get(record.levelname),
        Style.BRIGHT,
        "(%(levelname)1.1s)",
        Style.RESET_ALL,
        " ",
    ]
    asctime = [
        "[",
        Fore.BLACK,
        Style.DIM,
        Style.BRIGHT,
        "%(asctime)s",
        Fore.RESET,
        Style.RESET_ALL,
        "]",
    ]
    name = [
        " ",
        Fore.WHITE,
        Style.DIM,
        Style.BRIGHT,
        "%(name)s",
        Fore.RESET,
        Style.RESET_ALL,
        " ",
    ]
    module_funcName = [Fore.GREEN, Style.BRIGHT, "%(module)s.%(funcName)s()"]
    lineno = [
        Fore.BLACK,
        Style.DIM,
        Style.BRIGHT,
        ":",
        Style.RESET_ALL,
        Fore.CYAN,
        "%(lineno)d",
    ]

    tpl = "".join(reset + levelname + asctime + name + module_funcName + lineno + reset)

    return tpl


class DebugLogFormatter(LogFormatter):

    """Provides greater technical details than standard log Formatter."""

    template = debug_log_template



import pytest

from cihai import exc


def test_base_exception():
    with pytest.raises(exc.CihaiException):
        raise exc.CihaiException()  # Make sure its base of CihaiException

    with pytest.raises(Exception):
        raise exc.CihaiException()  # Extends python base exception



#!/usr/bin/env python
"""Chinese characater decomposition.

cihai.datasets.decomp
~~~~~~~~~~~~~~~~~~~~~

See http://cjkdecomp.codeplex.com/wikipage?title=cjk-decomp&referringTitle=Home

The CJK Decomposition Data File is a graphical analysis of the approx 75,000
Chinese/Japanese characters in Unicode. The latest version is 0.4.0, updated on
15 August 2012.

The data comprises the 36 strokes (0x31c0..0x31e3), the 115 radicals
(0x2e80..0x2ef3, except 0x2e9a), the 20924 unified characters (0x4e00..0x9fbb),
the 12 unique characters from the compatibility range (0xf900..0xfad9), the
6582 extension A characters (0x3400..0x4db5), the 42711 extension B characters
(0x20000..0x2a6d6), the 4149 extension C characters (0x2a700..0x2b734), and the
222 extension D characters (0x2b740..0x2b81d).

Each record has 3 fields, viz, the character being defined, the type of
decomposition, and a list of zero or more constituent components, like so:

的:a(白,勺)

The character being defined and the constituent components are either a Unihan
token, in the basic or a supplemental plane, or a 5-digit number representing
an intermediate decomposition not in Unicode. There are approx 10,000 such
intermediate decompositions.

If you need a font, you can use the Hanazono font.

Only pictorial configurations are used, not semantic ones. Where characters
have typeface differences I've used the one provided by the Mainland Chinese
contribution to Unicode. When there's more than one possible configuration,
I've selected one only.

The possible configurations and their meanings are:
Code regex       Meaning         Number of possible constituents
c        component       0
m.*      modified in some way, e.g. me=equivalent, msp=special, mo=outline, \
         ml=left radical version      1
w.*      second constituent contained within first in some way, e.g. w=within \
         at the center, wbl=within at bottom left   2
ba|d     second between first moving across or downwards         2
lock     components locked together      2
s.*      first component surrounds second, e.g. s=surrounds fully, \
         str=surrounds around the top-right    2
a        flows across    >= 2
d        flows downwards         >= 2
r.*      repeats and/or reflects in some way, e.g. refh=reflect horizontally, \
         rot=rotate 180 degrees, rrefr= repeat with a reflection rightwards, \
         ra=repeat across, r3d=repeat 3 times downwards, r3tr=repeat in a \
         triangle, rst=repeat surrounding around the top       1


The s, a, d, and r codes may be followed by /t, /m, /s, or /o, to show whether
the join touches, molds, snaps together, or overlaps, respectively.

Some more work needs to be done, including reducing the quantity of
intermediate components by removing duplicates, lowering the number of
components in many sequences, reanalysis of decomposition configurations, and
of course quality checking and corrections.

Last edited Dec 6, 2013 at 8:09 PM by gavingrover, version 3

"""
import logging

__copyright__ = "Copyright 2013-2018 Tony Narlock."
__license__ = "MIT, see LICENSE for details."

log = logging.getLogger(__name__)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['cihai', 'cihai.data', 'cihai.data.decomp', 'cihai.data.unihan']

package_data = \
{'': ['*']}

install_requires = \
['appdirs', 'kaptan', 'sqlalchemy<1.4', 'unihan-etl>=0.18.0,<0.19.0']

setup_kwargs = {
    'name': 'cihai',
    'version': '0.17.1a0',
    'description': 'Library for CJK (chinese, japanese, korean) language data.',
    'long_description': '# cihai &middot; [![Python Package](https://img.shields.io/pypi/v/cihai.svg)](https://pypi.org/project/cihai/) [![License](https://img.shields.io/github/license/cihai/cihai.svg)](https://github.com/cihai/cihai/blob/master/LICENSE) [![Code Coverage](https://codecov.io/gh/cihai/cihai/branch/master/graph/badge.svg)](https://codecov.io/gh/cihai/cihai)\n\nPython library for [CJK](https://cihai.git-pull.com/glossary.html#term-cjk) (chinese, japanese,\nkorean) data.\n\nThis project is under active development. Follow our progress and check back for updates!\n\n## Quickstart\n\n### API / Library (this repository)\n\n```console\n$ pip install --user cihai\n```\n\n```python\nfrom cihai.core import Cihai\n\nc = Cihai()\n\nif not c.unihan.is_bootstrapped:  # download and install Unihan to db\n    c.unihan.bootstrap(unihan_options)\n\nquery = c.unihan.lookup_char(\'好\')\nglyph = query.first()\nprint("lookup for 好: %s" % glyph.kDefinition)\n# lookup for 好: good, excellent, fine; well\n\nquery = c.unihan.reverse_char(\'good\')\nprint(\'matches for "good": %s \' % \', \'.join([glph.char for glph in query]))\n# matches for "good": 㑘, 㑤, 㓛, 㘬, 㙉, 㚃, 㚒, 㚥, 㛦, 㜴, 㜺, 㝖, 㤛, 㦝, ...\n```\n\nSee [API](https://cihai.git-pull.com/api.html) documentation and\n[/examples](https://github.com/cihai/cihai/tree/master/examples).\n\n### CLI ([cihai-cli](https://cihai-cli.git-pull.com))\n\n```console\n$ pip install --user cihai-cli\n```\n\nCharacter lookup:\n\n```console\n$ cihai info 好\n```\n\n```yaml\nchar: 好\nkCantonese: hou2 hou3\nkDefinition: good, excellent, fine; well\nkHangul: 호\nkJapaneseOn: KOU\nkKorean: HO\nkMandarin: hǎo\nkTang: "*xɑ̀u *xɑ̌u"\nkTotalStrokes: "6"\nkVietnamese: háo\nucn: U+597D\n```\n\nReverse lookup:\n\n```console\n$ cihai reverse library\n```\n\n```yaml\nchar: 圕\nkCangjie: WLGA\nkCantonese: syu1\nkCihaiT: \'308.302\'\nkDefinition: library\nkMandarin: tú\nkTotalStrokes: \'13\'\nucn: U+5715\n--------\n```\n\n### UNIHAN data\n\nAll datasets that cihai uses have stand-alone tools to export their data. No library required.\n\n- [unihan-etl](https://unihan-etl.git-pull.com) - [UNIHAN](http://unicode.org/charts/unihan.html)\n  data exports for csv, yaml and json.\n\n## Developing\n\n```console\n$ git clone https://github.com/cihai/cihai.git`\n```\n\n```console\n$ cd cihai/\n```\n\n[Bootstrap your environment and learn more about contributing](https://cihai.git-pull.com/contributing/). We use the same conventions / tools across all cihai projects: `pytest`, `sphinx`, `flake8`, `mypy`, `black`, `isort`, `tmuxp`, and file watcher helpers (e.g. `entr(1)`).\n\n## Quick links\n\n- [Quickstart](https://cihai.git-pull.com/quickstart.html)\n- [Datasets](https://cihai.git-pull.com/datasets.html) a full list of current and future data sets\n- Python [API](https://cihai.git-pull.com/api.html)\n- [Roadmap](https://cihai.git-pull.com/design-and-planning/)\n- Python support: >= 3.7, pypy\n- Source: <https://github.com/cihai/cihai>\n- Docs: <https://cihai.git-pull.com>\n- Changelog: <https://cihai.git-pull.com/history.html>\n- API: <https://cihai.git-pull.com/api.html>\n- Issues: <https://github.com/cihai/cihai/issues>\n- Test coverage: <https://codecov.io/gh/cihai/cihai>\n- pypi: <https://pypi.python.org/pypi/cihai>\n- OpenHub: <https://www.openhub.net/p/cihai>\n- License: MIT\n\n[![Docs](https://github.com/cihai/cihai/workflows/docs/badge.svg)](https://cihai.git-pull.com/)\n[![Build Status](https://github.com/cihai/cihai/workflows/tests/badge.svg)](https://github.com/cihai/cihai/actions?query=workflow%3A%22tests%22)\n',
    'author': 'Tony Narlock',
    'author_email': 'tony@git-pull.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://cihai.git-pull.com',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



__title__ = "cihai"
__package_name__ = "cihai"
__version__ = "0.17.1a0"
__description__ = "Library for CJK (chinese, japanese, korean) language data."
__author__ = "Tony Narlock"
__email__ = "tony@git-pull.com"
__github__ = "https://github.com/cihai/cihai"
__docs__ = "https://cihai.git-pull.com"
__tracker__ = "https://github.com/cihai/cihai/issues"
__pypi__ = "https://pypi.org/project/cihai/"
__license__ = "MIT"
__copyright__ = "Copyright 2013- cihai software foundation (Tony Narlock)"



# -*- coding: utf8 - *-
from cihai.core import Cihai
from cihai.data.unihan import bootstrap


def test_reflect_db(tmpdb_file, unihan_options):
    c = Cihai(
        {"database": {"url": "sqlite:///{tmpdb_file}".format(tmpdb_file=tmpdb_file)}}
    )
    assert not c.unihan.is_bootstrapped
    bootstrap.bootstrap_unihan(c.sql.metadata, unihan_options)
    assert not hasattr(c.sql.base.classes, "Unihan")
    c.unihan.sql.reflect_db()
    assert hasattr(c.sql.base.classes, "Unihan")



#!/usr/bin/env python
from cihai.core import Cihai
from cihai.extend import Dataset

data = {}  # any data source, internal, a file, on the internet, in a database...


class MyDataset(Dataset):
    def bootstrap(self):  # automatically ran with .add_dataset, if exists
        # Use this to setup your dataset, check if updates are needed, etc.
        data.update({"好": "Good", "好好": "Hello"})

    def givemedata(self, key):
        return data[key]

    def search(self, needle):
        return {k: v for k, v in data.items() if needle in k}

    def backwards(self, needle):
        return [k for k, v in data.items() if needle in v]


def run():
    c = Cihai(unihan=False)

    c.add_dataset(MyDataset, namespace="moo")
    c.moo.bootstrap()

    print("Definitions exactly for 好", c.moo.givemedata("好"))

    print("Definitions matching with 你好:", ", ".join(c.moo.search("好")))

    print("Reverse definition with Good:", ", ".join(c.moo.backwards("Good")))


if __name__ == "__main__":
    run()
